#ifndef lint
static char *RCSid() { return RCSid("$Id: amavis.c,v 1.1.2.10 2001/03/12 14:16:46 lhecking Exp $"); }
#endif

/*
 * Add some copyright notice here ...
 *
 * Usage: amavis sender recipient [recipient ...] [-- lda [lda-args]]
 *
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#define BUFFLEN 8192
#define DIRBUFFLEN 108
#include <stdio.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>
#include <time.h>

#define TEMPLATE "/amavis-XXXXXXXX"

static void mydebug(const char *, ...);
static char *mymktempdir(char *);

static int debugflag;
static char *mydebugfile;

#ifndef HAVE_STRLCPY
# define strlcpy mystrlcpy
static size_t
mystrlcpy(char *dst, const char *src, size_t dstsize)
{
    (void) strncpy(dst, src, dstsize - 1);
    *(dst + dstsize - 1) = '\0';
    return strlen(src);
}
#endif

static void
mydebug(const char *fmt, ...)
{
    FILE *f;
    time_t tmpt;
    char *mystime;
    va_list ap;

    if (debugflag) {

	if (mydebugfile) {
	    f = fopen(mydebugfile, "a");
	} else {
	    f = fopen(RUNTIME_DIR "/virdebug", "a");
	}
	tmpt = time(NULL);
	mystime = ctime(&tmpt);
	fprintf(f, "%ld:%s", (long) getpid(), mystime);

	va_start(ap, fmt);
	vfprintf(f, fmt, ap);
	va_end(ap);

	fputc('\n', f);
	fclose(f);
    }
}


static char *
mymktempdir(char *s)
{
    char dirname[DIRBUFFLEN];
    char *stt;
    int count = 0;

    strlcpy(dirname, s, sizeof(dirname));

#ifdef HAVE_MKDTEMP
    return mkdtemp(s);
#else
    /* magic number alert */
    while (count++ < 20) {
# ifdef HAVE_MKTEMP
	stt = mktemp(s);
# else
	stt = strchr(dirname, '-') + 1;
	if (stt) {
	    /* more magic number alert */
	    snprintf(stt, sizeof(dirname) - 1 - (stt - dirname), "%08d", lrand48() / 215);
	    /* This assumes that s in the calling function is the same size
	     * as the local dirname! Beware of malicious coders ;-) */
	    strlcpy(s, dirname, sizeof(dirname));
	    stt = dirname;
	} else {
	    /* invalid template */
	    return NULL;
	}
# endif
	if (stt) {
	    if (!mkdir(s, S_IRWXU)) {
		return s;
	    } else {
		continue;
	    }
	}
    }
    return NULL;
#endif /* HAVE_MKDTEMP */
}

/* Simple "protocol" */
const char _LDA = '\2';
const char _EOT = '\3';

/* take input from stdin as an email, with argv as sender and recips
 * then pass it all to amavisd
 * TODO: sender/recipient parsing for qmail; see qmail-queue(8) for details
 */
int
main(int argc, char **argv)
{
    char atmpfile[BUFFLEN];
    char buff[BUFFLEN];
    char dirname[DIRBUFFLEN];
    char xstat[8] = { 0 };
    struct sockaddr_un saddr;
    FILE *fout;
    int r, sock, i = 2, lda = 0, msgsize = 0;
    char retval;
    struct stat StatBuf;
#if !(HAVE_MKDTEMP || HAVE_MKTEMP)
    int mypid = getpid();

    srand48(time(NULL) ^ (mypid + (mypid << 15)));
#endif

    /* Process args first */
    if (argc < 3) {
	perror("Insufficient number of arguments");
	exit(EX_TEMPFAIL);
    }

    umask(0077);

    /* Find out whether an lda was specified (sendmail/local setup) */
    while (++i < argc) {
	if (strcmp(argv[i], "--") == 0) {
	    if (i < argc - 1)
		lda = i + 1;
	    /* once we've located the position of "--", we're done */
	    break;
	}
    }

    /* */
    strlcpy(dirname, RUNTIME_DIR, sizeof(dirname));
    strncat(dirname, TEMPLATE, sizeof(dirname) - 1 - strlen(dirname));
    if (mymktempdir(dirname) == NULL) {
	fprintf(stderr, "Failed to create temp dir\n");
	exit(EX_TEMPFAIL);
    }

    if (lstat(dirname, &StatBuf) < 0) {
	fprintf(stderr, "%s: Error while trying lstat(%s)\n", argv[0], dirname);
	exit(EX_TEMPFAIL);
    }

    /* may be too restrictive for you, but's good to avoid problems */
    if (!S_ISDIR(StatBuf.st_mode) || StatBuf.st_uid != getuid() || StatBuf.st_gid != getgid()
	|| !(StatBuf.st_mode & (S_IWUSR | S_IRUSR))) {
	fprintf(stderr,
		"%s: Security Warning: %s must be a Directory and owned by User %d and Group %d\n"
		"and just read-/write-able by the User and noone else. Exit.\n", argv[0], dirname, getuid(), getgid());
	exit(EX_TEMPFAIL);
    }
    /* there is still a race condition here if RUNTIME_DIR is writeable by the attacker :-\ */

    snprintf(atmpfile, sizeof(dirname), "%s/email.txt", dirname);
    fout = fopen(atmpfile, "w");
    while (!feof(stdin)) {
	r = fread(buff, sizeof(char), sizeof(buff), stdin);
	fwrite(buff, sizeof(char), r, fout);
	msgsize += r;
    }
    fclose(fout);
    mydebug("size=%s\n", msgsize);

    r = (sock = socket(PF_UNIX, SOCK_STREAM, 0));
    if (r < 0) {
	perror("failed to allocate socket");
    }
    saddr.sun_family = AF_UNIX;
    strlcpy(saddr.sun_path, AMAVISD_SOCKET, sizeof(saddr.sun_path));
    if (r >= 0) {
	mydebug("connect()\n");
	r = connect(sock, (struct sockaddr *) &saddr, sizeof(saddr));
	if (r < 0)
	    perror("failed to connect()");
    }
    if (r >= 0) {
	mydebug("senddir()\n");
	r = send(sock, dirname, strlen(dirname), 0);
	if (r < 0)
	    perror("failed to send() directory");
    }
    if (r >= 0) {
	r = recv(sock, &retval, 1, 0);
	if (r < 0)
	    perror("failed to recv() directory confirmation");
    }
    if (r >= 0) {
	mydebug("sendfrom() %s\n", argv[1]);

	r = send(sock, argv[1], strlen(argv[1]), 0);
	if (r < 0)
	    perror("failed to send() Sender info");
    }
    if (r >= 0) {
	r = recv(sock, &retval, 1, 0);
	if (r < 0)
	    perror("failed to recv() ok for Sender info");
    }
    if (r >= 0) {
	for (i = 2; i < argc; i++) {
	    if (lda == 0) {
		/* No LDA - this branch is essentially the old code */
		mydebug("sendto() %s\n", argv[i]);
		r = send(sock, argv[i], strlen(argv[i]), 0);
		if (r < 0) {
		    perror("failed to send() recip info");
		} else {
		    r = recv(sock, &retval, 1, 0);
		    if (r < 0)
			perror("failed to recv() ok for recip info");
		}
	    } else {		/* lda != 0 */
		/* LDA specified */
		if (lda == i + 1) {
		    mydebug("sendlda() %s\n", argv[i]);
		    r = send(sock, &_LDA, 1, 0);
		} else {
		    mydebug("sendlda/arg() %s\n", argv[i]);
		    r = send(sock, argv[i], strlen(argv[i]), 0);
		}
		if (r < 0) {
		    perror("failed to send/lda() recip info");
		} else {
		    r = recv(sock, &retval, 1, 0);
		    if (r < 0)
			perror("failed to recv() ok for recip/lda info");
		}

	    }			/* else lda != 0 */
	}
    }
    if (r >= 0) {
	mydebug("sendEOT()\n");
	r = send(sock, &_EOT, 1, 0);
	/* send "end of args" msg */
	if (r < 0) {
	    perror("failed to send() EOT");
	} else {
	    r = recv(sock, xstat, 6, 0);
	    mydebug("received %s from daemon", xstat);
	    if (r < 0)
		perror("Failed to recv() final result");
	    /* get back final result */
	}
    }
    close(sock);
    mydebug("finished conversation\n");
    if (r < 0) {
	/* housekeeping */
	unlink(atmpfile);
	rmdir(dirname);
	mydebug("failing with EX_TEMPFAIL\n");
	exit(EX_TEMPFAIL);
	/* some point of the communication failed miserably - so give up */
    }
    retval = atoi(xstat);
    mydebug("%ld: retval is %d\n", (long) getpid(), retval);

    if (retval)
	exit(EX_TEMPFAIL);

    exit(retval);
}
